#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "asdf/asdf.h"
#include "asdf/import.h"

#include "util/region.h"

ASDF* import_vol(const char* filename, int ni, int nj, int nk,
                 const float offset, const float mm_per_voxel)
{

    // Add an empty boundary
    ni += 2;
    nj += 2;
    nk += 2;

    Region r = (Region){.imin = 0,  .jmin = 0,  .kmin = 0,
                        .ni = ni-1, .nj = nj-1, .nk = nk-1,
                        .voxels = (ni-1)*(nj-1)*(nk-1)};
    
    float dx = mm_per_voxel*ni,
          dy = mm_per_voxel*nj,
          dz = mm_per_voxel*nk;
    build_arrays(&r, 0, 0, 0, dx, dy, dz);
    
    // Load ALL THE DATA into a bunch of arrays allocated on the heap.
    // This will break on large file sizes (since it grows as n^3)
    FILE* file = fopen(filename, "rb");
    float*** data = malloc(sizeof(float**)*nk);
    for (int k=0; k < nk; ++k) {
        data[k] = malloc(sizeof(float*)*nj);
        for (int j=0; j < nj; ++j) {
            data[k][j] = malloc(sizeof(float*)*ni);
            for (int i=0; i < ni; ++i) {
            
                if (i == 0 || i == ni - 1 ||
                    j == 0 || j == nj - 1 ||
                    k == 0 || k == nk - 1)
                {
                    data[k][j][i] = 1;
                } else {
                    fscanf(file, "%4c", (char*)&data[k][j][i]);
                    data[k][j][i] = (-data[k][j][i]) + offset;
                }
            }
        }
    }
    fclose(file);
    
    ASDF* asdf = build_from_vol(data, r);
    
    free_arrays(&r);
    for (int k=0; k < nk; ++k) {
        for (int j=0; j < nj; ++j) {
            free(data[k][j]);
        }
        free(data[k]);
    }
    free(data);
    
    return asdf;
}

ASDF* import_vol_region(const char* filename, int ni, int nj, int nk,
                        const Region r, const int shift, const float offset)
{
    // Create a downsampled region on which to create the tree.
    Region r_ = (Region){ .imin = 0, .jmin = 0, .kmin = 0,
                          .ni = (r.ni - 1) >> shift,
                          .nj = (r.nj - 1) >> shift,
                          .nk = (r.nk - 1) >> shift};
    build_arrays(&r_, r.X[0],    r.Y[0],    r.Z[0],
                      r.X[r.ni] - r.X[0],
                      r.Y[r.nj] - r.Y[0],
                      r.Z[r.nk] - r.Z[0]);
    
    // Allocate space for resampled data
    float*** data = malloc(sizeof(float**)*(r_.nk+1));
    for (int k=0; k <= r_.nk; ++k) {
        data[k] = malloc(sizeof(float*)*(r_.nj+1));
        for (int j=0; j <= r_.nj; ++j) {
            data[k][j] = calloc(r_.ni+1, sizeof(float*));
        }
    }


    // Load region of data into the resampled data array
    FILE* file = fopen(filename, "rb");
    for (int k=r.kmin; k < r.kmin+r.nk; k += (1 << shift)) {
        for (int j=r.jmin; j < r.jmin+r.nj; j += (1 << shift)) {
        
            // Seek to the start of this data block
            fseek(file, 4*(k*ni*nj + j*ni + r.imin), SEEK_SET);

            for (int i=r.imin; i < r.imin+r.ni; i += (1 << shift)) {
                     
                float sample;
                fscanf(file, "%4c", (char*)&sample);
                
                float* target = &(data[(k-r.kmin)>>shift]
                                      [(j-r.jmin)>>shift]
                                      [(i-r.imin)>>shift]);
                *target = -sample + offset;
                
                // Slide forward, skipping unsampled values
                for (int c=0; c < 4*((1<<shift)-1); ++c) {
                    fgetc(file);
                }
                
            }
        }
    }
    fclose(file);    

    // Eat the data and return an ASDF
    ASDF* asdf = build_from_vol(data, r_);
    
    // FREE ALL THE DATA!
    for (int k=0; k <= r_.nk; ++k) {
        for (int j=0; j <= r_.nj; ++j) {
            free(data[k][j]);
        }
        free(data[k]);
    }
    free(data);
    free_arrays(&r_);
    
    return asdf;
}



ASDF* import_vol_resampled(const char* filename, int ni, int nj, int nk,
                           const int shift, const float offset,
                           const float mm_per_voxel)
{

    float dx = mm_per_voxel*ni,
          dy = mm_per_voxel*nj,
          dz = mm_per_voxel*nk;

    Region r = (Region){.imin = 0,  .jmin = 0,  .kmin = 0,
                        .ni = (ni-1) >> shift,
                        .nj = (nj-1) >> shift,
                        .nk = (nk-1) >> shift};
    r.voxels = r.ni*r.nj*r.nk;
    
    build_arrays(&r, 0, 0, 0, dx, dy, dz);
    
    // Allocate space for resampled data
    float*** data = malloc(sizeof(float**)*(r.nk+1));
    for (int k=0; k <= r.nk; ++k) {
        data[k] = malloc(sizeof(float*)*(r.nj+1));
        for (int j=0; j <= r.nj; ++j) {
            data[k][j] = calloc(r.ni+1, sizeof(float*));
        }
    }
    
    // Load the data into an array
    FILE* file = fopen(filename, "rb");
    printf("Reading data from file:\n");
    for (int k=0; k < nk; ++k) {
        printf("\t%i/%i\n", k, nk);
        for (int j=0; j < nj; ++j) {
            for (int i=0; i < ni; ++i) {
                     
                float* target = &(data[k>>shift][j>>shift][i>>shift]);
                float sample;
                fscanf(file, "%4c", (char*)&sample);
                
                if ( ((i >> shift) << shift) != i ||
                     ((j >> shift) << shift) != j ||
                     ((k >> shift) << shift) != k) continue;
                *target = -sample + offset;
            }
        }
    }
    fclose(file);    

    ASDF* asdf = build_from_vol(data, r);
    
    free_arrays(&r);
    for (int k=0; k <= r.nk; ++k) {
        for (int j=0; j <= r.nj; ++j) {
            free(data[k][j]);
        }
        free(data[k]);
    }
    free(data);
    
    return asdf;
}



ASDF* build_from_vol(float*** data, Region region)
{
    ASDF* asdf = calloc(1, sizeof(ASDF));
    
    if (region.voxels == 1) {
        // Pull x, y, and z values from the region
        *asdf = (ASDF) {
            .state = LEAF,
            .X = (Interval){region.X[0], region.X[region.ni]},
            .Y = (Interval){region.Y[0], region.Y[region.nj]},
            .Z = (Interval){region.Z[0], region.Z[region.nk]}
        };
    
        int count = 0;
        for (int n=0; n < 8; ++n) {
            int i = region.imin + ((n & 4) ? region.ni : 0),
                j = region.jmin + ((n & 2) ? region.nj : 0),
                k = region.kmin + ((n & 1) ? region.nk : 0);
            asdf->d[n] = data[k][j][i];
            
            if (asdf->d[n] < 0) count++;
            else                count--;
        }
        
        if (count == 8)         asdf->state = FILLED;
        else if (count == -8)   asdf->state = EMPTY;
        else                    asdf->state = LEAF;
        
    } else {
        asdf->state = BRANCH;
        
        Region octants[8];
        uint8_t bits = octsect(region, octants);
        for (int i=0; i < 8; ++i) {
            if (bits & (1<<i)) {
                asdf->branches[i] = build_from_vol(data, octants[i]);
            }
        }
        
        // Pull d[i] and X, Y, Z values from children nodes
        get_d_from_children(asdf);
        get_xyz_from_children(asdf);
        
        // Merge cells if possible
        simplify(asdf);
    }
    
    return asdf;
}