#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "asdf/asdf.h"
#include "asdf/mesh.h"
#include "asdf/cache.h"

#include "util/region.h"
#include "util/macros.h"

const uint8_t VERTEX_LOOP[] = {6, 4, 5, 1, 3, 2, 6};

// Based on which vertices are filled, this map tells you which
// edges to interpolate between when forming zero, one, or two
// triangles.
// (filled vertex is first in the pair)
const int EDGE_MAP[16][2][3][2] = {
    {{{-1,-1}, {-1,-1}, {-1,-1}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // ----
    {{{ 0, 2}, { 0, 1}, { 0, 3}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // ---0
    {{{ 1, 0}, { 1, 2}, { 1, 3}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // --1-
    {{{ 1, 2}, { 1, 3}, { 0, 3}}, {{ 0, 3}, { 0, 2}, { 1, 2}}}, // --10
    {{{ 2, 0}, { 2, 3}, { 2, 1}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // -2--
    {{{ 0, 3}, { 2, 3}, { 2, 1}}, {{ 2, 1}, { 0, 1}, { 0, 3}}}, // -2-0
    {{{ 1, 0}, { 2, 0}, { 2, 3}}, {{ 2, 3}, { 1, 3}, { 1, 0}}}, // -21-
    {{{ 2, 3}, { 1, 3}, { 0, 3}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // -210

    {{{ 3, 0}, { 3, 1}, { 3, 2}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // 3---
    {{{ 3, 2}, { 0, 2}, { 0, 1}}, {{ 0, 1}, { 3, 1}, { 3, 2}}}, // 3--0
    {{{ 1, 2}, { 3, 2}, { 3, 0}}, {{ 3, 0}, { 1, 0}, { 1, 2}}}, // 3-1-
    {{{ 1, 2}, { 3, 2}, { 0, 2}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // 3-10
    {{{ 3, 0}, { 3, 1}, { 2, 1}}, {{ 2, 1}, { 2, 0}, { 3, 0}}}, // 32--
    {{{ 3, 1}, { 2, 1}, { 0, 1}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // 32-0
    {{{ 3, 0}, { 1, 0}, { 2, 0}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // 321-
    {{{-1,-1}, {-1,-1}, {-1,-1}}, {{-1,-1}, {-1,-1}, {-1,-1}}}, // 3210
};

////////////////////////////////////////////////////////////////////////////////
 
int count_leafs(const ASDF* const asdf)
{
    if (!asdf) {
        return 0;
    } else if (asdf->state == LEAF) {
        return 1;
    } else if (asdf->state == BRANCH) {
        int count = 0;
        for (int i=0; i < 8; ++i)
            count += count_leafs(asdf->branches[i]);
        return count;
    } else {
        return 0;
    }
}

int count_tris(const ASDF* const asdf)
{
    if (!asdf) {
        return 0;
    } else if (asdf->state == LEAF) {
        
        // Loop over the six tetrahedra that make up a voxel cell
        int count = 0;
        for (int t = 0; t < 6; ++t) {
            // Find vertex positions for this tetrahedron
            uint8_t vertices[] = {0, 7, VERTEX_LOOP[t], VERTEX_LOOP[t+1]};
            uint8_t lookup = 0;
            for (int vertex = 3; vertex >= 0; --vertex) {
                lookup = (lookup << 1) + (asdf->d[vertices[vertex]] < 0);
            }
            if (EDGE_MAP[lookup][0][0][0] != -1) count++;
            if (EDGE_MAP[lookup][1][0][0] != -1) count++;
        }
        return count;
        
    } else if (asdf->state == BRANCH) {
        int count = 0;
        for (int i=0; i < 8; ++i)
            count += count_tris(asdf->branches[i]);
        return count;
    } else {
        return 0;
    }
}

////////////////////////////////////////////////////////////////////////////////

void find_bounds(const float* const p, const int count,
                 Vec3f* const min, Vec3f* const max)
{
    *min = (Vec3f){ INFINITY,  INFINITY,  INFINITY};
    *max = (Vec3f){-INFINITY, -INFINITY, -INFINITY};
    
    for (int i=0; i < count; i += 6) {
        if (p[i] < min->x)   min->x = p[i];
        if (p[i] > max->x)   max->x = p[i];

        if (p[i+1] < min->y)   min->y = p[i+1];
        if (p[i+1] > max->y)   max->y = p[i+1];

        if (p[i+2] < min->z)   min->z = p[i+2];
        if (p[i+2] > max->z)   max->z = p[i+2];
    }
}

////////////////////////////////////////////////////////////////////////////////

float find_radius(const float* const p, int count, Vec3f center)
{
    float radius2 = -INFINITY;
    
    for (int i=0; i < count; i += 6) {
        float r2 = pow(center.x - p[i], 2) +
                   pow(center.y - p[i+1], 2) +
                   pow(center.z - p[i+2], 2);
        
        if (r2 > radius2)
            radius2 = r2;
    }

    return sqrt(radius2);
}

////////////////////////////////////////////////////////////////////////////////

void triangulate(const ASDF* const asdf,
                 float* const vdata, uint32_t* const vcount,
                 uint32_t* const idata, uint32_t* const icount,
                 int* const halt)
{
    int ni, nj, nk;
    find_dimensions(asdf, &ni, &nj, &nk);
    
    
    Region r = (Region){.imin = 0,  .jmin = 0,  .kmin = 0,
                        .ni   = ni, .nj   = nj, .nk   = nk,
                        .voxels = ni*nj*nk
                        };
                        
    Edge* cache     = calloc(1, sizeof(Edge));
    cache->valid = false;
    Edge* subcache  = edge_subcache(cache,
                                    r.imin, r.imin + r.ni,
                                    r.jmin, r.jmin + r.nj,
                                    r.kmin, r.kmin + r.nk);
    
    *vcount = 0;
    _triangulate(asdf, r, subcache, vdata, vcount, idata, icount, halt);
    
    free_edge_cache(cache);
}   

////////////////////////////////////////////////////////////////////////////////

void _triangulate(const ASDF* const asdf, const Region r, Edge* const cache,
                  float* const vdata, uint32_t* const vcount,
                  uint32_t* const idata, uint32_t* const icount,
                  int* const halt)
{
    if (*halt || !asdf) return;

    if (asdf->state == LEAF) {
        evaluate_voxel(asdf, r, cache, vdata, vcount, idata, icount);
    } else if (asdf->state == BRANCH) {
    
        // Split the region and recurse down the octree.
        Region octants[8];
        octsect_merged(r, asdf, octants);
    
        Edge* subcache  = edge_subcache(cache,
                                        r.imin, r.imin + r.ni,
                                        r.jmin, r.jmin + r.nj,
                                        r.kmin, r.kmin + r.nk);
    
        for (int i=0; i < 8; ++i)
            _triangulate(asdf->branches[i], octants[i], subcache,
                         vdata, vcount, idata, icount, halt);
    }
}

////////////////////////////////////////////////////////////////////////////////

void evaluate_voxel(const ASDF* const leaf, const Region r, Edge* const cache,
                    float* const vdata, uint32_t* const vcount,
                    uint32_t* const idata, uint32_t* const icount)
{
    // Store a cache of interpolated zero crossings on all
    // possible edges of the cube.
    Edge* edge_cache[8][8];
    for (int i=0; i < 8; ++i)
        for (int j=0; j < 8; ++j)
            edge_cache[i][j] = NULL;

    // Loop over the six tetrahedra that make up a voxel cell
    for (int t = 0; t < 6; ++t) {
    
        // Find vertex positions for this tetrahedron
        uint8_t vertices[] = {0, 7, VERTEX_LOOP[t], VERTEX_LOOP[t+1]};
        
        // Figure out which of the sixteen possible combinations
        // we're currently experiencing.
        uint8_t lookup = 0;
        for (int vertex = 3; vertex >= 0; --vertex)
            lookup = (lookup << 1) + (leaf->d[vertices[vertex]] < 0);
        
        if (EDGE_MAP[lookup][0][0][0] == -1)    continue;

        // Store the first triangle in the vertex array
        for (int v=0; v < 3; ++v) {        
            idata[(*icount)++] = zero_crossing(leaf, r, edge_cache, cache,
                                      vertices[EDGE_MAP[lookup][0][v][0]],
                                      vertices[EDGE_MAP[lookup][0][v][1]],
                                      vdata, vcount);
        }
        
        // Move on to the second triangle, aborting if there isn't one.
        if (EDGE_MAP[lookup][1][0][0] == -1)    continue;
        
        // Store the second triangle in the vertex array
        for (int v=0; v < 3; ++v) {        
            idata[(*icount)++] = zero_crossing(leaf, r, edge_cache, cache,
                                      vertices[EDGE_MAP[lookup][1][v][0]],
                                      vertices[EDGE_MAP[lookup][1][v][1]],
                                      vdata, vcount);
        }
    }
}

////////////////////////////////////////////////////////////////////////////////

uint32_t zero_crossing(const ASDF* const leaf, const Region r,
                       Edge* edge_cache[8][8], Edge* const cache,
                       const uint8_t v0, const uint8_t v1,
                       float* const vdata, uint32_t* const vcount)
{
    // If we've already solved for this zero crossing, return the answer.
    if (edge_cache[v0][v1]) return edge_cache[v0][v1]->id;

    // Look up this position in the global cache
    uint16_t i1 = (v0 & 4) ? r.imin + r.ni : r.imin,
             j1 = (v0 & 2) ? r.jmin + r.nj : r.jmin,
             k1 = (v0 & 1) ? r.kmin + r.nk : r.kmin;
    uint16_t i2 = (v1 & 4) ? r.imin + r.ni : r.imin,
             j2 = (v1 & 2) ? r.jmin + r.nj : r.jmin,
             k2 = (v1 & 1) ? r.kmin + r.nk : r.kmin;

    Edge* e = get_edge(cache, i1, j1, k1, i2, j2, k2);
    edge_cache[v0][v1] = e;
    edge_cache[v1][v0] = e;
    
    // If this edge isn't already solved, find the zero crossing
    // using binary search.
    if (!e->valid) {
        float x0 = (v0 & 4) ? leaf->X.upper : leaf->X.lower,
              y0 = (v0 & 2) ? leaf->Y.upper : leaf->Y.lower,
              z0 = (v0 & 1) ? leaf->Z.upper : leaf->Z.lower;
    
        float dx = ((v1 & 4) ? leaf->X.upper : leaf->X.lower) - x0,
              dy = ((v1 & 2) ? leaf->Y.upper : leaf->Y.lower) - y0,
              dz = ((v1 & 1) ? leaf->Z.upper : leaf->Z.lower) - z0;
    
        float d = 0.5;
        float step = 0.25;
        float result;
        
        // Binary search along the edge to find the zero crossing
        for (int iteration = 0; iteration < 32; ++iteration) {
        
            // Use interpolation to find the distance metric value at this edge
            result = asdf_interpolate(leaf, x0 + d*dx, y0 + d*dy, z0 + d*dz);
        
            // Change the binary search coefficient accordingly
            if (result < -EPSILON)         d += step;
            else if (result > EPSILON)     d -= step;
            else                        break;
            
            // And shrink the step size
            step /= 2;
        }
        
        e->pos = (Vec3f){x0 + d*dx, y0 + d*dy, z0 + d*dz};
        e->gradient = (Vec3f){0, 0, 0};
        
        e->valid = true;
        e->id = (*vcount)/6;
        
        vdata[(*vcount)++] = e->pos.x;
        vdata[(*vcount)++] = e->pos.y;
        vdata[(*vcount)++] = e->pos.z;
        vdata[(*vcount)++] = e->gradient.x;
        vdata[(*vcount)++] = e->gradient.y;
        vdata[(*vcount)++] = e->gradient.z;
    }

    // The gradient will be a sum from all of the cells
    // that share this vertex.
    Vec3f g = asdf_gradient(leaf, e->pos.x, e->pos.y, e->pos.z);
    vdata[e->id*6+3] += g.x;
    vdata[e->id*6+4] += g.y;
    vdata[e->id*6+5] += g.z;
    
    return e->id;
}
        

void increase_indices(uint32_t* const idata, uint32_t icount, uint32_t di)
{
    for (uint32_t i = 0; i < icount; ++i)
        idata[i] += di;
}

////////////////////////////////////////////////////////////////////////////////

void write_stl(const uint32_t* const idata, const uint32_t icount,
               const float* const vdata, const char* filename)
{
    FILE* stl = fopen(filename, "wb");
    
    fprintf(stl, "This is a binary STL file created by asdf_stl. Learn more at kokompe.cba.mit.edu");
    
    uint32_t stl_faces = icount/3;
    for (int i=0; i<4; ++i) {
        fputc(((char*)&stl_faces)[i], stl);
    }

    int index=0;
    while (index < icount) {

        // Write the face normal (which we'll keep empty)
        for (int j=0; j < 12; ++j) fputc(0, stl);
        
        for (int vertex=0; vertex < 3; vertex++) {        
            float xyz[3] = {vdata[6*idata[index]],
                            vdata[6*idata[index]+1],
                            vdata[6*idata[index]+2]};
            for (int j=0; j < 12; ++j)
                fputc(((char*)&xyz)[j], stl);
            index++;
        }
        
        fputc(0, stl);
        fputc(0, stl);
    }
    
    fclose(stl);
}
        
////////////////////////////////////////////////////////////////////////////////

void write_mesh(const char* filename,
                const Interval X, const Interval Y, const Interval Z,
                const uint32_t* const idata, const uint32_t icount,
                const float* const vdata, const uint32_t vcount)
{
    FILE* mesh = fopen(filename, "wb");
    
    float header_bounds[] = {X.lower, X.upper,
                             Y.lower, Y.upper,
                             Z.lower, Z.upper};

    for (int h=0; h<6*4; ++h) {
        fputc(((char*)&header_bounds)[h], mesh);
    }

    for (int i=0; i < 4; ++i) {
        fputc(((char*)&icount)[i], mesh);
    }
    
    for (int v=0; v < 4; ++v) {
        fputc(((char*)&vcount)[v], mesh);
    }
    
    fwrite((void*)idata, sizeof(uint32_t), icount,   mesh);
    fwrite((void*)vdata, sizeof(float),    vcount*6, mesh);    
    
    fclose(mesh);
}
    
void read_mesh_header(const char* filename,
                      Interval* const X, Interval* const Y, Interval* const Z,
                      uint32_t* const icount, uint32_t* vcount)
{
    FILE* mesh = fopen(filename, "rb");
    
    fscanf(mesh, "%4c", (char*)(&(X->lower)));
    fscanf(mesh, "%4c", (char*)(&(X->upper)));
    fscanf(mesh, "%4c", (char*)(&(Y->lower)));
    fscanf(mesh, "%4c", (char*)(&(Y->upper)));
    fscanf(mesh, "%4c", (char*)(&(Z->lower)));
    fscanf(mesh, "%4c", (char*)(&(Z->upper)));
    
    fscanf(mesh, "%4c", (char*)icount);
    fscanf(mesh, "%4c", (char*)vcount);
    
    fclose(mesh);
}

void read_mesh(const char* filename,
               const uint32_t* const idata, const uint32_t icount,
               const float* const vdata, const uint32_t vcount)
{
    FILE* mesh = fopen(filename, "rb");
    
    // Skip the header
    fseek(mesh, 8*4, SEEK_SET);

    // Load all the data
    fread((void*)idata, sizeof(uint32_t), icount,   mesh);
    fread((void*)vdata, sizeof(float),    vcount*6, mesh);
    
    fclose(mesh);
}