#ifndef PACKED_H
#define PACKED_H

#include "tree.h"

/*  ustack
 *
 *  A simple FIFO stack of unsigned integers.
 */
typedef struct ustack_ {
    unsigned count;
    struct ustack_* next;
} ustack;

/*  PackedTree
 *
 *  A structure containing nodes organized by rank
 */
typedef struct PackedTree_ {
    struct Node_*** nodes;
    unsigned* active;
    ustack**  disabled;
    
    unsigned num_levels;
    struct Node_* head;
} PackedTree;


/*  make_packed
 *
 *  Takes in a well-formed, deduplicated MathTree.
 *  Returns a PackedTree with the same nodes.
 */
PackedTree* make_packed(struct MathTree_* tree);

/*  free_packed
 *
 *  Frees a packed tree.  Does not free nodes, since they should be
 *  pointing to the same nodes as the original MathTree.
 */
void free_packed(PackedTree* packed);

/*  disable_nodes
 *
 *  Travels down the tree, disabling nodes whose values will not matter
 *  upon further subdivision, e.g. if we perform the operation max(A, B)
 *  and A is strictly less than B, we can ignore A on this interval and smaller.
 *
 *  Nodes are disabled by swapping them to the back of their respective list
 *  and decrementing the active count (so that the evaluation loop doesn't
 *  touch them at all).
 *
 *  A count of nodes disabled on this past is stored on the top of the
 *  disabled stack.
 */
void disable_nodes(PackedTree* tree);


/*  enable_nodes
 *
 *  Enables nodes that were disabled on the most recent call to disable_nodes.
 *  Nodes are enabled by increasing the value of active[level][op], so that
 *  an evaluation continues further down the list.
 *  
 *  This function also pops the top value off of the disabled stack.
 */
void enable_nodes(PackedTree* tree);

#endif