#include <stdlib.h>

#include "tree/tree.h"

#include "node/node.h"
#include "node/printers.h"
#include "util/switches.h"

MathTree* new_tree(unsigned num_levels, unsigned num_constants)
{
    MathTree* tree = malloc(sizeof(MathTree));
    
    *tree = (MathTree){
        .nodes      = num_levels ?
                        calloc(num_levels, sizeof(Node**) * LAST_OP) : NULL,
        .active     = num_levels ?
                        calloc(num_levels, sizeof(unsigned[LAST_OP])) : NULL,
        
        .constants = num_constants ?
                        malloc(sizeof(Node*)*num_constants) : NULL,
                        
        .num_constants = num_constants,
        .head = NULL,
        .num_levels = num_levels,
    };
    
    return tree;
}
    

void free_tree(MathTree* tree)
{
    for (unsigned level=0; level < tree->num_levels; ++level) {
        for (unsigned op=0; op < LAST_OP; ++op) {
            for (unsigned n=0; n < tree->active[level][op]; ++n) {
                free(tree->nodes[level][op][n]);
            }
            free(tree->nodes[level][op]);
        }
    }
    
    for (unsigned c=0; c < tree->num_constants; ++c) {
        free(tree->constants[c]);
    }
    
    free(tree->nodes);
    free(tree->active);
    free(tree->constants);
    
    free(tree);
}


void print_tree(MathTree* tree)
{
    print_node(tree->head);
}


void fprint_tree(MathTree* tree, int fd)
{
    fprint_node(tree->head, fd);
}


void print_tree_verbose(MathTree* tree)
{
    for (unsigned level=0; level < tree->num_levels; ++level) {
        printf("Level %i:\n", level);
        for (unsigned op=0; op < LAST_OP; ++op) {
            if (tree->active[level][op] == 0)   continue;
            printf("  %s:\n", OPCODE_NAMES[op]);
            for (unsigned n=0; n < tree->active[level][op]; ++n) {
                printf("    ");
                print_node(tree->nodes[level][op][n]);
                printf("\n");
            }
        }
    }
    
    if (tree->num_constants) {
        if (tree->num_levels) printf("\n");
        printf("Constants:\n");
        for (unsigned c=0; c < tree->num_constants; ++c) {
            printf("    ");
            print_node(tree->constants[c]);
            printf("\n");
        }
    }
}

void fprint_tree_verbose(MathTree* tree, int fd)
{
    FILE* f = fdopen(fd, "w");

    for (unsigned level=0; level < tree->num_levels; ++level) {
        fprintf(f, "Level %i:\n", level);
        for (unsigned op=0; op < LAST_OP; ++op) {
            if (tree->active[level][op] == 0)   continue;
            fprintf(f, "  %s:\n", OPCODE_NAMES[op]);
            for (unsigned n=0; n < tree->active[level][op]; ++n) {
                fprintf(f, "    ");
                _fprint_node(tree->nodes[level][op][n], f);
                fprintf(f, "\n");
            }
        }
    }
    
    if (tree->num_constants) {
        if (tree->num_levels) fprintf(f, "\n");
        fprintf(f, "Constants:\n");
        for (unsigned c=0; c < tree->num_constants; ++c) {
            fprintf(f, "    ");
            _fprint_node(tree->constants[c], f);
            fprintf(f, "\n");
        }
    }
    
    fclose(f);
}

void mark(MathTree* tree)
{
    for (unsigned level=0; level < tree->num_levels; ++level) {
        for (unsigned op=0; op < LAST_OP; ++op) {
            for (unsigned n=0; n < tree->active[level][op]; ++n) {
                tree->nodes[level][op][n]->marked = true;
            }
        }
    }
}


void  unmark(MathTree* tree)
{
    for (unsigned level=0; level < tree->num_levels; ++level) {
        for (unsigned op=0; op < LAST_OP; ++op) {
            for (unsigned n=0; n < tree->active[level][op]; ++n) {
                tree->nodes[level][op][n]->marked = false;
            }
        }
    }
}


void deduplicate_tree(MathTree* tree)
{
    int constants_removed = deduplicate_constants(tree->constants,
                                                  tree->num_constants);

    unsigned (*removed)[LAST_OP] =  calloc(tree->num_levels,
                                           sizeof(unsigned[LAST_OP]));
                                          
    for (int level=0; level < tree->num_levels; ++level) {
        for (int op=0; op < LAST_OP; ++op) {
            removed[level][op] = deduplicate_row(tree->nodes[level][op],
                                                 tree->active[level][op]);
        }
    }
    
    // If the head was deduplicated, follow the arrows.
    while (tree->head->clone_address) {
        tree->head = tree->head->clone_address;
    }
    
    // Free extra constants and reallocate array
    for (int c = tree->num_constants - constants_removed;
             c < tree->num_constants; ++c) {
        free(tree->constants[c]);
    }
    tree->num_constants -= constants_removed;
    tree->constants = realloc(tree->constants,
                              tree->num_constants*sizeof(Node*));
                              
    // Free extra nodes and reallocate arrays
    for (int level=0; level < tree->num_levels; ++level) {
        for (int op=0; op < LAST_OP; ++op) {
            for (int n=tree->active[level][op] - removed[level][op];
                     n < tree->active[level][op]; ++n) {
                free(tree->nodes[level][op][n]);
            }
            tree->active[level][op] -= removed[level][op];
            tree->nodes[level][op] = realloc(tree->nodes[level][op],
                               sizeof(Node*)*tree->active[level][op]);
        }
    }
    free(removed);
}

int deduplicate_constants(Node** constants, unsigned num)
{
    int removed = 0;
    
    for (int a=0; a < num-removed; ++a) {
        constants[a]->clone_address = NULL;
        for (int b=a+1; b < num-removed; ++b) {
            if (constants[a]->results.f == constants[b]->results.f) {
                
                constants[a]->clone_address = constants[b];
                
                // Swap the duplicate node to the back of the list
                int back = num-(++removed);
                Node* tmp = constants[back];
                constants[back] = constants[a];
                constants[a--] = tmp;
                
                break;
            }
        }
    }    
    
    return removed;
}


int deduplicate_row(Node** nodes, unsigned num)
{
    // Update pointers to children nodes
    for (int i=0; i < num; ++i) {
        while (nodes[i]->lhs && nodes[i]->lhs->clone_address) {
            nodes[i]->lhs = nodes[i]->lhs->clone_address;
        }
        while (nodes[i]->rhs && nodes[i]->rhs->clone_address) {
            nodes[i]->rhs = nodes[i]->rhs->clone_address;
        }
    }

    int removed = 0;
    for (int a=0; a < num-removed; ++a) {
        nodes[a]->clone_address = NULL;
        for (int b=a+1; b < num-removed; ++b) {
            if (nodes[a]->lhs == nodes[b]->lhs &&
                nodes[a]->rhs == nodes[b]->rhs) {
                
                nodes[a]->clone_address = nodes[b];
                int back = num-(++removed);
                Node* tmp = nodes[back];
                nodes[back] = nodes[a];
                nodes[a--] = tmp;
                
                break;
            }
        }
    }

    return removed;
}

MathTree* nonary_tree(void)
{
    // Create a tree and allocated storage space for everything
    MathTree* tree = new_tree(1, 0);
    return tree;
}

MathTree* X_tree(void)
{
    MathTree* tree = nonary_tree();
    tree->nodes[0][OP_X]    = malloc(sizeof(Node*));
    tree->nodes[0][OP_X][0] = X_n();
    
    tree->active[0][OP_X] = 1;
    tree->head = tree->nodes[0][OP_X][0];
    
    return tree;
}

MathTree* Y_tree(void)
{
    MathTree* tree = nonary_tree();
    tree->nodes[0][OP_Y]    = malloc(sizeof(Node*));
    tree->nodes[0][OP_Y][0] = Y_n();
    
    tree->active[0][OP_Y] = 1;
    tree->head = tree->nodes[0][OP_Y][0];
    
    return tree;
}

MathTree* Z_tree(void)
{
    MathTree* tree = nonary_tree();
    tree->nodes[0][OP_Z]    = malloc(sizeof(Node*));
    tree->nodes[0][OP_Z][0] = Z_n();
    
    tree->active[0][OP_Z] = 1;
    tree->head = tree->nodes[0][OP_Z][0];
    
    return tree;
}

MathTree* constant_tree(float value)
{
    // Create a tree and allocated storage space for everything
    MathTree* tree = new_tree(0, 1);
    
    tree->constants[0] = constant_n(value);
    tree->head = tree->constants[0];
    
    return tree;
}

////////////////////////////////////////////////////////////////////////////////

MathTree* unary_tree(MathTree* arg, Node* head)
{
    MathTree* tree = NULL;
    
    if (head->constant) {
        tree = new_tree(0, 1);
        tree->constants[0] = head;
        tree->head = tree->constants[0];
        
        return tree;
    }
    
    tree = new_tree(arg->num_levels+1, arg->num_constants);


    // Clone the constants into the new tree.
    for (int c=0; c < arg->num_constants; ++c) {
        tree->constants[c] = clone_node(arg->constants[c]);
    }

    // Then clone all of the nodes into the new tree.
    // Iterate over levels
    for (int level=0; level < arg->num_levels; ++level) {
        // Iterate over opcodes
        for (int op=0; op < LAST_OP; ++op) {
        
            int active = arg->active[level][op];
            tree->active[level][op] = active;
            if (active)
                tree->nodes[level][op] = malloc(sizeof(Node*)*active);
            else
                tree->nodes[level][op] = NULL;
            
            // Clone each node in this level and opcode.
            for (int n=0; n < active; ++n) {
                Node* clone = clone_node(arg->nodes[level][op][n]);
                tree->nodes[level][op][n] = clone;
            }
        }
    }
    
    // Adjust the child pointers to the clones that were just made
    head->lhs = head->lhs->clone_address;
    
    tree->nodes[arg->num_levels][head->opcode] = malloc(sizeof(Node*));
    tree->nodes[arg->num_levels][head->opcode][0] = head;
    tree->active[arg->num_levels][head->opcode]++;
    tree->head = head;
    
    
    return tree;
}

MathTree* square_tree(MathTree* n)
{
    return unary_tree(n, square_n(n->head));
}

MathTree* sqrt_tree(MathTree* n)
{
    return unary_tree(n, sqrt_n(n->head));
}

MathTree* neg_tree(MathTree* n)
{
    return unary_tree(n, neg_n(n->head));
}
       
////////////////////////////////////////////////////////////////////////////////

MathTree* binary_tree(MathTree* lhs, MathTree* rhs, Node* head)
{
    MathTree* tree = NULL;
    
    // If the head is constant, then return a simple tree that
    // only contains the constant head and nothing else.
    if (head->constant) {
        tree = new_tree(0, 1);
        tree->constants[0] = head;
        tree->head = tree->constants[0];
        
        return tree;
    }
    
    
    int num_levels = lhs->num_levels > rhs->num_levels ?
                     lhs->num_levels : rhs->num_levels;
    int num_constants = lhs->num_constants + rhs->num_constants;
    
    // Create a new tree to store the resulting data structure
    tree = new_tree(num_levels+1, num_constants);

    // Clone the constants into the new tree.
    for (int c=0; c < lhs->num_constants; ++c) {
        tree->constants[c] = clone_node(lhs->constants[c]);
    }
    for (int c=0; c < rhs->num_constants; ++c) {
        tree->constants[c+lhs->num_constants] = 
            clone_node(rhs->constants[c]);
    }

    // Then clone all of the nodes into the new tree.
    // Iterate over levels
    for (int level=0; level < num_levels; ++level) {
        // Iterate over opcodes
        for (int op=0; op < LAST_OP; ++op) {
        
            int lhs_active = level < lhs->num_levels ?
                             lhs->active[level][op] : 0;
            int rhs_active = level < rhs->num_levels ?
                             rhs->active[level][op] : 0;

        
            int active = lhs_active + rhs_active;
            tree->active[level][op] = active;
            if (active)
                tree->nodes[level][op] = malloc(sizeof(Node*)*active);
            else
                tree->nodes[level][op] = NULL;
            
            // Clone each node in this level and opcode.
            for (int n=0; n < lhs_active; ++n) {
                Node* clone = clone_node(lhs->nodes[level][op][n]);
                tree->nodes[level][op][n] = clone;
            }
            for (int n=0; n < rhs_active; ++n) {
                Node* clone = clone_node(rhs->nodes[level][op][n]);
                tree->nodes[level][op][n+lhs_active] = clone;
            }
            
        }
    }
    
    // Adjust the child pointers to the clones that were just made
    head->lhs  = head->lhs->clone_address;
    head->rhs  = head->rhs->clone_address;
    
    tree->nodes[tree->num_levels-1][head->opcode] = malloc(sizeof(Node*));
    tree->nodes[tree->num_levels-1][head->opcode][0] = head;
    tree->active[tree->num_levels-1][head->opcode]++;
    tree->head = head;
    
#if DEDUPLICATE
    deduplicate_tree(tree);
#endif
    return tree;
}

MathTree* add_tree(MathTree* lhs, MathTree* rhs)
{
    if (lhs->head->constant && lhs->head->results.f == 0)
        return clone_tree(rhs);
    else if (rhs->head->constant && rhs->head->results.f == 0)
        return clone_tree(lhs);
    else
        return binary_tree(lhs, rhs, add_n(lhs->head, rhs->head));
}

MathTree* sub_tree(MathTree* lhs, MathTree* rhs)
{
    if (lhs->head->constant && lhs->head->results.f == 0)
        return neg_tree(rhs);
    else if (rhs->head->constant && rhs->head->results.f == 0)
        return clone_tree(lhs);
    else
        return binary_tree(lhs, rhs, sub_n(lhs->head, rhs->head));
}

MathTree* mul_tree(MathTree* lhs, MathTree* rhs)
{
    if (lhs->head->constant && lhs->head->results.f == 0)
        return constant_tree(0);
    else if (rhs->head->constant && rhs->head->results.f == 0)
        return constant_tree(0);
    else if (lhs->head->constant && lhs->head->results.f == 1)
        return clone_tree(rhs);
    else if (rhs->head->constant && rhs->head->results.f == 1)
        return clone_tree(lhs);
    else
        return binary_tree(lhs, rhs, mul_n(lhs->head, rhs->head));
}

MathTree* div_tree(MathTree* lhs, MathTree* rhs)
{
    if (rhs->head->constant && rhs->head->results.f == 1)
        return clone_tree(lhs);
    else
        return binary_tree(lhs, rhs, div_n(lhs->head, rhs->head));
}

MathTree* min_tree(MathTree* lhs, MathTree* rhs)
{
    return binary_tree(lhs, rhs, min_n(lhs->head, rhs->head));
}

MathTree* max_tree(MathTree* lhs, MathTree* rhs)
{
    return binary_tree(lhs, rhs, max_n(lhs->head, rhs->head));
}

MathTree* pow_tree(MathTree* lhs, MathTree* rhs)
{
    if (rhs->head->constant && rhs->head->results.f == 0)
        return constant_tree(1);
    else if (rhs->head->constant && rhs->head->results.f == 1)
        return clone_tree(lhs);
    else if (rhs->head->constant && rhs->head->results.f == 2)
        return square_tree(lhs);
    else
        return binary_tree(lhs, rhs, pow_n(lhs->head, rhs->head));
}

////////////////////////////////////////////////////////////////////////////////
MathTree* clone_tree(MathTree* orig)
{
    MathTree* clone = new_tree(orig->num_levels, orig->num_constants);

    // Clone the constants into the new tree.
    for (int c=0; c < orig->num_constants; ++c) {
        clone->constants[c] = clone_node(orig->constants[c]);
    }

    // Then clone all of the nodes into the new tree:
    
    // Iterate over levels
    for (int level=0; level < orig->num_levels; ++level) {
        // Iterate over opcodes
        for (int op=0; op < LAST_OP; ++op) {
        
            int active = orig->active[level][op];
            clone->active[level][op] = active;
            if (active) {
                clone->nodes[level][op] = malloc(sizeof(Node*)*active);
            } else {
                clone->nodes[level][op] = NULL;
            }
            
            // Clone each node in this level and opcode.
            for (int n=0; n < active; ++n) {
                Node* clone_ = clone_node(orig->nodes[level][op][n]);
                clone->nodes[level][op][n] = clone_;
            }
        }
    }
    
    clone->head = orig->head->clone_address;
    return clone;
}
////////////////////////////////////////////////////////////////////////////////


MathTree* map_tree(MathTree* target, MathTree* X, MathTree* Y, MathTree* Z)
{
    // First, disable any maps that don't apply to this tree.
    _Bool X_found = false, Y_found = false, Z_found = false;
    if (target->num_levels > 0) {
        if (target->active[0][OP_X]) X_found = true;
        if (target->active[0][OP_Y]) Y_found = true;
        if (target->active[0][OP_Z]) Z_found = true;
    }
    if (!X_found)   X = NULL;
    if (!Y_found)   Y = NULL;
    if (!Z_found)   Z = NULL;

    int num_levels = target->num_levels;
    if (X && X->num_levels > num_levels) num_levels = X->num_levels;
    if (Y && Y->num_levels > num_levels) num_levels = Y->num_levels;
    if (Z && Z->num_levels > num_levels) num_levels = Z->num_levels;
    
    int num_constants = target->num_constants +
                        (X ? X->num_constants : 0) + 
                        (Y ? Y->num_constants : 0) + 
                        (Z ? Z->num_constants : 0);
                        
    // First, we'll clone all of the nodes into a tree.
    MathTree* tree = new_tree(num_levels, num_constants);

    // Clone the constants into the new tree.
    int constant_index = 0;
    for (int c=0; c < target->num_constants; ++c) {
        tree->constants[constant_index++] = clone_node(target->constants[c]);
    }
    for (int c=0; X && c < X->num_constants; ++c) {
        tree->constants[constant_index++] = clone_node(X->constants[c]);
    }
    for (int c=0; Y && c < Y->num_constants; ++c) {
        tree->constants[constant_index++] = clone_node(Y->constants[c]);
    }
    for (int c=0; Z && c < Z->num_constants; ++c) {
        tree->constants[constant_index++] = clone_node(Z->constants[c]);
    }


    // First, we'll construct a tree where mapped nodes stay on the same level.
    // This means that they'll end up in the wrong position by rank, which
    // will be corrected later on in the process.
    for (int level=0; level < num_levels; ++level) {
        for (int op=0; op < LAST_OP; ++op) {
            
            // Count the number of nodes in this level+opcode across
            // all three expressions.
            int target_active = level < target->num_levels ?
                                target->active[level][op] : 0;
            int active = target_active;
            if (X && level < X->num_levels) active += X->active[level][op];
            if (Y && level < Y->num_levels) active += Y->active[level][op];
            if (Z && level < Z->num_levels) active += Z->active[level][op];
        
            // Let this point to the end of the currently populated array
            // so that we can later insert the nodes from 'target' at the
            // back of the array.
            tree->active[level][op] = active - target_active;
            if (active)
                tree->nodes[level][op] = malloc(sizeof(Node*)*active);
            else
                tree->nodes[level][op] = NULL;
            
            // Clone each node in this level and opcode.
            int node_index = 0;
            if (X && level < X->num_levels) {
                for (int n=0; n < X->active[level][op]; ++n) {
                    Node* clone = clone_node(X->nodes[level][op][n]);
                    tree->nodes[level][op][node_index++] = clone;
                }
            }
            if (Y && level < Y->num_levels) {
                for (int n=0; n < Y->active[level][op]; ++n) {
                    Node* clone = clone_node(Y->nodes[level][op][n]);
                    tree->nodes[level][op][node_index++] = clone;
                }
            }
            if (Z && level < Z->num_levels) {
                for (int n=0; n < Z->active[level][op]; ++n) {
                    Node* clone = clone_node(Z->nodes[level][op][n]);
                    tree->nodes[level][op][node_index++] = clone;
                }
            }
        }
    }
    
    // Now, copy all of the nodes in the mapped expression into the tree
    // (with special cases for X, Y, and Z nodes, which may get remapped to
    //  the head of the cloned replacement expression).
    for (int level=0; level < target->num_levels; ++level) {
        // Iterate over opcodes
        for (int op=0; op < LAST_OP; ++op) {
            for (int n=0; n < target->active[level][op]; ++n) {
                Node* node = target->nodes[level][op][n];
                if (X && op == OP_X) {
                    node->clone_address = X->head->clone_address;
                } else if (Y && op == OP_Y) {
                    node->clone_address = Y->head->clone_address;
                } else if (Z && op == OP_Z) {
                    node->clone_address = Z->head->clone_address;
                } else {
                    Node* clone = clone_node(node);
                    int index = tree->active[level][op]++;
                    tree->nodes[level][op][index] = clone;
                }
            }
        }
    }
    tree->head = target->head->clone_address;
    
    // Shuffle up the tree so that nodes end up at the right heights
    rejigger_tree(tree);
    
    // Remove duplicates from the tree
#if DEDUPLICATE
    deduplicate_tree(tree);
#endif

    return tree;
}

void count_nodes(MathTree* tree, unsigned (*count)[LAST_OP])
{
    for (int level=0; level < tree->num_levels; ++level) {
        for (int op=0; op < LAST_OP; ++op) {
            for (int n=0; n < tree->active[level][op]; ++n) {
                Node* node = tree->nodes[level][op][n];
                count[node->rank][node->opcode]++;
            }
        }
    }
}

void rejigger_tree(MathTree* tree)
{

    // Find the highest ranked node in the tree.
    // This will determine the size of many allocated arrays
    int max_rank = -1;
    for (int level=0; level < tree->num_levels; ++level) {
        for (int op=0; op < LAST_OP; ++op) {
            for (int n=0; n < tree->active[level][op]; ++n) {
                if (tree->nodes[level][op][n]->rank > max_rank) {
                    max_rank = tree->nodes[level][op][n]->rank;
                }
            }
        }
    }
    int num_levels = max_rank + 1;
    
    // Figure out how many nodes there are of each rank and opcode.
    unsigned (*count)[LAST_OP]  = calloc(num_levels, sizeof(unsigned[LAST_OP]));
    count_nodes(tree, count);

    // Create new arrays for nodes and active counts.
    Node **((*nodes)[LAST_OP]);
    if (num_levels) nodes = calloc(num_levels, sizeof(Node**) * LAST_OP);
    else            nodes = NULL;
    
    unsigned (*active)[LAST_OP];
    if (num_levels) active = calloc(num_levels, sizeof(unsigned[LAST_OP]));
    else            active = NULL;
    
    // Allocate space for all of the new nodes.
    for (int level=0; level < num_levels; ++level) {
        for (int op=0; op < LAST_OP; ++op) {
            if (count[level][op])
                nodes[level][op] = malloc(sizeof(Node*)*count[level][op]);
            else
                nodes[level][op] = NULL;
        }
    }
    free(count);
    
    
    // And move them over, placing them in the level appropriate for
    // their rank (which may be different than their current level)
    for (int level=0; level < tree->num_levels; ++level) {
        for (int op=0; op < LAST_OP; ++op) {
            for (int n=0; n < tree->active[level][op]; ++n) {
                Node* node = tree->nodes[level][op][n];
                int index = active[node->rank][op]++;
                nodes[node->rank][op][index] = node;
            }
            free(tree->nodes[level][op]);
        }
    }
    
    free(tree->nodes);
    free(tree->active);
    
    tree->nodes      = nodes;
    tree->active     = active;
    tree->num_levels = num_levels;
}