import ctypes
from datetime       import datetime
import threading
from math           import sin, cos, radians, log, ceil
import os
import Queue

from interval       import Interval
from multithread    import multithread

class _ASDF(ctypes.Structure):
    ''' ctypes data structure containing an ASDF. '''
    pass
_ASDF._fields_ = [('state', ctypes.c_int),
                  ('X', Interval), ('Y', Interval), ('Z', Interval),
                  ('branches', ctypes.POINTER(_ASDF)*8),
                  ('d', ctypes.c_float*8)]

# Imports that require _ASDF to be defined
from libtree     import libtree
from region      import Region
from vec3f       import Vec3f
from pixels      import PixelBuffer
from mesh        import Mesh

################################################################################

class ASDF(object):
    ''' Wrapper class that contains an ASDF pointer and
        automatically frees it upon destruction.'''
        
    def __init__(self, ptr, free=True):
        ''' Create an ASDF from a pointer.
        
            The pointer is freed upon destruction if free is True.
        '''
        self.ptr = ptr
        self.free = free
        self._scale = None
    
    def __del__(self):
        if self.free and libtree is not None:
            libtree.free_asdf(self.ptr)
    
    def interpolate(self, x, y, z):
        return libtree.interpolate(
                        self.ptr.contents.d, x, y, z,
                        self.X.lower, self.Y.lower, self.Z.lower,
                        self.X.upper - self.X.lower,
                        self.Y.lower - self.Y.upper,
                        self.Z.lower - self.Z.upper)
    
    def make_extrapolation(self):
        libtree.make_extrapolation(self.ptr)
    
    def set_depth(self, d):
        libtree.set_depth(self.ptr, d)
    
    def clear_depth(self):
        libtree.clear_depth(self.ptr)
    
    @property
    def branches(self):
        b = []
        for i in range(8):
            try:
                self.ptr.contents.branches[i].contents.X
            except ValueError:
                b += [None]
            else:
                b += [ASDF(self.ptr.contents.branches[i], free=False)]
        return b

    @property
    def state(self):
        return ['FILLED','EMPTY','BRANCH','LEAF'][self.ptr.contents.state]
    @property
    def d(self): return [self.ptr.contents.d[i] for i in range(8)]
    @property
    def X(self): return self.ptr.contents.X
    @property
    def Y(self): return self.ptr.contents.Y
    @property
    def Z(self): return self.ptr.contents.Z
    
    @property
    def scale(self): return self._scale
    @scale.setter
    def scale(self, value):
        value = float(value)
        if self._scale is None:
            libtree.asdf_scale(self.ptr, value)
        else:
            libtree.asdf_scale(self.ptr, value/self._scale)
        self._scale = float(value)
        return self._scale
    
    @property
    def depth(self):    return libtree.get_depth(self.ptr)
    
    @property
    def dimensions(self):
        ni = ctypes.c_int()
        nj = ctypes.c_int()
        nk = ctypes.c_int()
        
        libtree.find_dimensions(self.ptr, ni, nj, nk)
        return ni.value, nj.value, nk.value
        
    
    def save(self, filename):
        if self.scale is None:
            print 'Warning: saving ASDF with unknown scale'
            print '(defaulting to 1 mm/unit)'
        libtree.asdf_write(self.ptr, filename)


    @classmethod
    def load(cls, filename):
        ''' Loads an asdf file. '''
        asdf = cls(libtree.asdf_read(filename))
        asdf.scale = 1
        return asdf
    
    
    @classmethod
    def from_vol(cls, filename, ni, nj, nk, offset, mm_per_voxel, shift=None):
        '''Imports a .vol file.'''
        if shift is None:
            asdf = cls(libtree.import_vol(filename, ni, nj, nk,
                                          offset, mm_per_voxel))
        else:
            asdf = cls(libtree.import_vol_resampled(filename, ni, nj, nk,
                                        shift, offset, mm_per_voxel))
        asdf.scale = 1
        return asdf
    
    @staticmethod
    def construct(tree, region, mm_per_unit=None, interrupt=None):
        ''' Constructs an ASDF from a math tree.
        
            'interrupt' should be a threading.Event to
            halt evaluation partway through.
        '''
        # Create an event to interrupt the evaluation
        if interrupt is None:   interrupt = threading.Event()
        
        # Shared flag to interrupt rendering
        halt = ctypes.c_int(0)
        
        # Split the region into up to 8 sections
        subregions = (Region*8)()
        subregions = region.octsect(all=True)
        subregions = [(i, subregions[i]) for i in range(8)
                                         if subregions[i] is not None]

        threads = len(subregions)        
        clones  = [tree.clone() for i in range(threads)]
        packed  = [libtree.make_packed(c.ptr) for c in clones]

        # Generate a root for the tree
        asdf = ASDF(libtree.asdf_root(packed[0], region))
        
        # Multithread the solver process
        q = Queue.Queue()
        args = zip(packed, subregions, [q]*threads, [halt]*threads)
        multithread(ASDF.construct_branch, args, interrupt, halt)

        # Attach the branches to the root
        for s in subregions:
            try:
                id, branch = q.get_nowait()
            except Queue.Empty:
                break
            else:
                asdf.ptr.contents.branches[id] = branch
        libtree.get_d_from_children(asdf.ptr)
        
        # Set a scale on the ASDF if one was provided
        if mm_per_unit is not None:     asdf.scale = mm_per_unit
        
        return asdf
           
    @staticmethod
    def construct_branch(tree, (id, region), queue, halt):
        ''' Constructs an ASDF and pushes it to the queue.
            
            Frees the tree after construction is complete.
        '''
        asdf = libtree.build_asdf(tree, region, halt)
        queue.put((id, asdf))
        libtree.free_packed(tree)
        
    
    def bounds(self, alpha=0, beta=0):
        corners = [Vec3f(self.X.upper if (i & 4) else self.X.lower,
                         self.Y.upper if (i & 2) else self.Y.lower,
                         self.Z.upper if (i & 1) else self.Z.lower)
                   for i in range(8)]
        M = (ctypes.c_float*4)(cos(radians(alpha)), sin(radians(alpha)),
                               cos(radians(beta)),  sin(radians(beta)))
        corners = [libtree.project(c, M) for c in corners]
        
        xmin = min(c.x for c in corners)
        xmax = max(c.x for c in corners)
        ymin = min(c.y for c in corners)
        ymax = max(c.y for c in corners)
        zmin = min(c.z for c in corners)
        zmax = max(c.z for c in corners)
        
        return (xmin, ymin, zmin, xmax, ymax, zmax)
    
    def bounding_region(self, resolution, alpha=0, beta=0):
        b = self.bounds(alpha, beta)
        return Region(b[0:3], b[3:6], resolution)
    
    
    #
    #   Render to image
    #
    def render(self, region=None, threads=8, alpha=0, beta=0, resolution=10):
        ''' Renders an ASDF on a given region, return a PixelBuffer image.
            Resolution is in pixels/mm, since ASDF is always scaled
            to millimeter units.'''
        
        return self.render_multi(region, threads, alpha, beta, resolution)[0]
    
    
    def render_multi(self, region=None, threads=8,
                     alpha=0, beta=0, resolution=10):
        ''' Renders an ASDF on a given region, returning three images.
            Resolution is in pixels/mm, since ASDF is always scaled
            to millimeter units.'''
        
        if region is None:
            region = self.bounding_region(resolution, alpha, beta)
            
        depth   = PixelBuffer(region.ni, region.nj, channels=1, depth=16)
        shaded  = PixelBuffer(region.ni, region.nj, channels=1, depth=16)
        normals = PixelBuffer(region.ni, region.nj, channels=3, depth=8)
        
        subregions = region.split_xy(threads)
        
        M = (ctypes.c_float*4)(cos(radians(alpha)), sin(radians(alpha)),
                               cos(radians(beta)), sin(radians(beta)))
        args = [(self.ptr, s, M, depth.pixels, shaded.pixels, normals.pixels)
                for s in subregions]
        multithread(libtree.render_asdf, args)
        
        for image in [depth, shaded, normals]:
            image.bounds = (ctypes.c_float*6)(*self.bounds(alpha, beta))
        return depth, shaded, normals
    
    #
    #   Triangulation functions
    #           
    def triangulate(self, interrupt=None):

        # Create an event to interrupt the evaluation
        if interrupt is None:   interrupt = threading.Event()
        
        # Shared flag to interrupt rendering
        halt = ctypes.c_int(0)

        q = Queue.Queue()
        args = [(b, halt, q) for b in self.branches if b is not None]
        multithread(ASDF._triangulate, args, interrupt, halt)

        meshes = []
        while True:
            try:                m = q.get_nowait()
            except Queue.Empty: break
            else:               meshes.append(m)
                
        mesh = Mesh.merge(meshes)
        
        return mesh
        
    def _triangulate(self, halt, queue=None):
        ''' Triangulate, using recursion to avoid creating
            buffers that are unreasonably large.
            
            Returns a mesh object.'''
            
        # Estimate the number of MB of RAM it will take to triangulate
        # this ASDF subtree.
        count = libtree.count_leafs(self.ptr)
        if count == 0:  return None
        MB = 3*(count*6*2*3*6*4 + count*6*2*3*4) / 1e6
        
        # If this subtree is small, solve it
        if MB < 250:
        
            # Count the number of triangles more precisely
            count = libtree.count_tris(self.ptr)
            
            # Allocate space for vertices and indexes
            vdata  = (ctypes.c_float*(count*3*6))()
            vcount =  ctypes.c_uint32()
            
            idata  = (ctypes.c_uint32*(count*3))()
            icount =  ctypes.c_uint32()

            libtree.triangulate(self.ptr, vdata, vcount, idata, icount, halt)
            vdata_trimmed = (ctypes.c_float*vcount.value)()
            
            ctypes.memmove(vdata_trimmed, vdata, vcount.value*4)
            
            mesh = Mesh(vdata_trimmed, idata,
                        Interval(self.X.lower, self.X.upper),
                        Interval(self.Y.lower, self.Y.upper),
                        Interval(self.Z.lower, self.Z.upper))
            if queue is None:
                return mesh
            else:
                queue.put(mesh)
        
        # Otherwise, recurse and solve octants
        else:
            meshes = [b._triangulate(halt) for b in self.branches]
            meshes = filter(lambda m: m is not None, meshes)
            mesh = Mesh.merge(meshes)
            
            if queue is None:   return mesh
            else:               queue.put(mesh)