import ctypes
import wx

from libtree import libtree

class PixelBuffer(object):
    ''' Class to wrap a C-style pixel array (indexed by row, column). '''
    
    def __init__(self, w, h, channels=1, depth=8):
    
        # The buffer is the actual memory location of the pixels
        if depth == 8:
            base_type = ctypes.c_ubyte
        elif depth == 16:
            base_type = ctypes.c_uint16
        else:
            raise ValueError('Invalid bit depth (must be 8 or 16)')
            
        self.buffer = (base_type * (w*h*channels))()
        
        # The pixel array points to either single bytes (for a one-channel
        # image) or triples of bytes (for a three-channel image)
        if channels == 1:
            self.pixels = (ctypes.POINTER(base_type) * h)()
            ptype = ctypes.POINTER(base_type)
        elif channels == 3:
            self.pixels = (ctypes.POINTER(base_type*3) * h)()
            ptype = ctypes.POINTER(base_type*3)
        else:
            raise ValueError('Invalid number of channels (must be 1 or 3)')
            
        start = ctypes.addressof(self.buffer)


        # We create a 2D array of pointers representing individual pixels
        for j in range(h):
            self.pixels[h-j-1] = ctypes.cast(start + w*j*channels*depth/8,
                                             ptype)
            
        self.width  = w
        self.height = h
        self._channels = channels
        self._depth = depth
        self.bounds = (ctypes.c_float*6)(0, 0, 0, 1, 1, 1)
        self._wx = None
    
    @property
    def wximg(self):
        ''' Returns a wx.Image of this Image.'''
        if self._wx:    return self._wx
        
        w, h = self.width, self.height
        if self.channels == 1:
            white = chr(255)*w*h*3
            self._wx = wx.ImageFromBuffer(w, h, white, self.buffer)
        elif self.channels == 3:
            self._wx = wx.ImageFromBuffer(w, h, self.buffer)
        
        return self._wx
    
    @property
    def channels(self):
        return self._channels
    @channels.setter
    def channels(self, value):
        if value == self.channels:
            return self.channels
        if self.depth == 8 and self.channels == 1:
            if value != 3:
                raise ValueError('Invalid target channel count %i' % value)
            new = PixelBuffer(self.width, self.height, channels=3)
            libtree.triple_image(self.width, self.height,
                                 self.pixels, new.pixels)
            self._channels = 3
            self.pixels = new.pixels
            self.buffer = new.buffer
            return self.channels
        else:
            raise ValueError('Cannot modify channels of %-bit, %-channel image'
                             % (self.depth, self.channels))
    
    @property
    def depth(self):
        return self._depth
    @depth.setter
    def depth(self, value):
        if value == self.depth:
            return self.depth
        if self.depth == 16 and self.channels == 1:
            if value != 8:
                raise ValueError('Invalid target depth %i' % value)
            new = PixelBuffer(self.width, self.height)
            libtree.downsample_image(self.width, self.height,
                                     self.pixels, new.pixels)
            self._depth = 8
            self.pixels = new.pixels
            self.buffer = new.buffer
            return self.depth
        else:
            raise ValueError('Cannot modify depth of %-bit, %-channel image'
                             % (self.depth, self.channels))
    
    def flipped_pixels(self):
        if self.depth == 8:
            flip = (ctypes.POINTER(ctypes.c_uint8) * self.height)()
        elif self.depth == 16:
            flip = (ctypes.POINTER(ctypes.c_uint16) * self.height)()
        
        for i in range(self.height):
            flip[i] = self.pixels[self.height-i-1]
        return flip
    
    
    def save(self, filename):
        if filename[-4:].lower() != '.png':
            raise TypeError('Image must be saved with .png extension')
        
        if self.depth == 8:
            self.wximg.SaveFile(filename, wx.BITMAP_TYPE_PNG)
        elif self.depth == 16:
            libtree.save_png16L(filename, self.width, self.height,
                                self.bounds, self.flipped_pixels())