from math import pi

import wx
from wx import glcanvas

from koko.c.libtree import libtree
from koko.c.vec3f   import Vec3f

try:
    import OpenGL
    from OpenGL.GL      import *
    from OpenGL.arrays  import vbo
    from OpenGL.GL      import shaders
    GL_enabled = True
except ImportError:
    GL_enabled = False

################################################################################

class GLCanvas(glcanvas.GLCanvas):
    def __init__(self, parent, size=(500, 300)):
        glcanvas.GLCanvas.__init__(self, parent, wx.ID_ANY, size=size)
        self.context = glcanvas.GLContext(self)
        self.init = False
        self.size = None
        
        self.Bind(wx.EVT_SIZE,      self.evt_size)
        self.Bind(wx.EVT_PAINT,     self.evt_paint)
        
        self.Bind(wx.EVT_LEFT_DOWN, self.evt_mouse_left_down)
        self.Bind(wx.EVT_LEFT_UP,   self.evt_mouse_left_up)
        self.Bind(wx.EVT_MOTION,    self.evt_mouse_move)
        self.Bind(wx.EVT_MOUSEWHEEL, self.evt_mouse_scroll)
        
        self.mouse_spin = self.mouse_pan = False
        self.alpha = 0
        self.beta  = 0
        
        self.scale = 1
        self.center = Vec3f(0, 0, 0)

        # Select from plain, wire, or normal shaders
        self.shader     = 'plain'
        self.index_vbo  = None
        
        self.path_vbo   = None
        self.path       = None
       
################################################################################
       
    def clear(self):
        self.index_vbo  = None
        self.path_vbo   = None
        self.path       = None
        self.Refresh()

    def clear_path(self):
        self.path_vbo   = None
        self.path       = None
        self.Refresh()
        
################################################################################
        
    @property
    def border(self):
        try:
            return self._border
        except AttributeError:
            self._border = None
            return self._border
    @border.setter
    def border(self, value):
        self._border = value
        wx.CallAfter(self.Refresh)
        
    
    def get_pixels(self):
        width, height = self.GetClientSize() 
        glReadBuffer(GL_FRONT)
        glPixelStorei(GL_PACK_ALIGNMENT, 1) 
        return width, height, glReadPixels(2, 2, width,height, 
                                           GL_RGB, GL_UNSIGNED_BYTE)
################################################################################
        
    def snap_axis(self, axis):
        if axis == '+x':
            self.alpha, self.beta = pi/2, 0
        elif axis == '+y':
            self.alpha, self.beta = 0, pi/2
        elif axis == '+z':
            self.alpha, self.beta = 0, 0
        elif axis == '-x':
            self.alpha, self.beta = -pi/2, 0
        elif axis == '-y':
            self.alpha, self.beta = 0, -pi/2
        elif axis == '-z':
            self.alpha, self.beta = pi, 0
        self.Refresh()
       
################################################################################
       
    def evt_size(self, evt=None):
        if not self.init:
            self.init_GL()
        if self.init:
            self.update_viewport()
    
################################################################################
    
    def evt_paint(self, evt=None):
        if not self.init:
            self.init_GL()
            self.update_viewport()
        if self.init:
            self.draw()

################################################################################

    def evt_mouse_left_down(self, evt):
        self.mouse = wx.Point(evt.GetX(), evt.GetY())
        if wx.GetKeyState(wx.WXK_SHIFT):
            self.mouse_pan  = True
        else:
            self.mouse_spin = True

################################################################################
    
    def evt_mouse_left_up(self, evt):
        self.mouse_spin = self.mouse_pan = False
    
################################################################################
    
    def evt_mouse_move(self, evt):
        pos = wx.Point(evt.GetX(), evt.GetY())
        if self.mouse_spin:
            delta = pos - self.mouse
            self.alpha = (self.alpha + delta.x) % 360
            self.beta  -= delta.y
            if   self.beta < 0: self.beta = 0
            elif self.beta > 180: self.beta = 180
            self.Refresh()
        elif self.mouse_pan:
            delta = pos - self.mouse
            delta = Vec3f(0.01/self.scale*delta.x, -0.01/self.scale*delta.y, 0)
            proj = delta.deproject(self.alpha, self.beta)
            self.center.x -= proj.x
            self.center.y -= proj.y
            self.center.z -= proj.z
            self.Refresh()
        self.mouse = pos    

################################################################################

    def evt_mouse_scroll(self, evt):
        if wx.GetKeyState(wx.WXK_SHIFT):
            delta = Vec3f(0, 0, 0.01*evt.GetWheelRotation()/self.scale)
            proj = delta.deproject(self.alpha, self.beta)
            self.center.x -= proj.x
            self.center.y -= proj.y
            self.center.z -= proj.z
        else:
            dScale = 1 / 1.0025 if evt.GetWheelRotation() < 0 else 1.0025
            for i in range(abs(evt.GetWheelRotation())):
                self.scale *= dScale
        self.Refresh()

################################################################################

    def update_viewport(self):
        self.SetCurrent(self.context)
        w, h = self.Size
        glViewport(0, 0, w, h)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity();
        
        aspect = w/float(h)

        if aspect > 1:  glFrustum(-aspect, aspect, -1, 1, 1, 9)
        else:           glFrustum(-1, 1, -1/aspect, 1/aspect, 1, 9)
            
        glMatrixMode(GL_MODELVIEW)
    
################################################################################
    
    def compile_shaders(self):

        vshader = """
            #version 120
            attribute vec3 vertex_position;
            attribute vec3 vertex_normal;
            
            varying vec3 normal;
            
            void main() {
                gl_Position = gl_ModelViewProjectionMatrix *
                              vec4(vertex_position, 1.0);
                              
                normal = normalize(gl_NormalMatrix*vertex_normal);
            }
        """

        path_shader = """
            #version 120
            attribute vec3 vertex_position;
            attribute float vertex_color;
            
            varying float color;
            
            void main() {
                gl_Position = gl_ModelViewProjectionMatrix *
                              vec4(vertex_position, 1.0);
                color = vertex_color;
            }
        """
        
        frag_path = """
            #version 120
            
            varying float color;
            
            void main() {
                gl_FragColor = vec4(0.1, 0.9, 0.25, color);
            }
        """
        
        cube_vert = """
            #version 120
            attribute vec3 vertex_position;
            attribute vec3 vertex_color;
            
            varying vec3 color;
            
            void main() {
                gl_Position = gl_ModelViewProjectionMatrix *
                              vec4(vertex_position, 1.0);
                color = vertex_color;
            }
        """
        
        cube_frag = """
            #version 120
            
            varying vec3 color;
            
            void main() {
                gl_FragColor = vec4(color[0], color[1], color[2], 1 );
            }
        """
        
        frag_norm = """
            #version 120
            
            varying vec3 normal;
            
            void main() {
                gl_FragColor = vec4(normal[0]/2+0.5,
                                    normal[1]/2+0.5,
                                    normal[2]/2+0.5, 1 );
            }
        """
        frag_wire = """
            #version 120
            
            varying vec3 normal;
            
            void main() {
                float B = normal[2] < 0 ? 0.2 : normal[2]*0.8+0.2;
                gl_FragColor = vec4(B, B, B, 1 );
            }
        """
        frag_shaded = """
            #version 120
            
            varying vec3 normal;
            
            void main() {
                gl_FragColor = vec4(normal[2]*0.8, normal[2]*0.8, normal[2]*0.8, 1 );
            }
        """
        self.plain_shader = shaders.compileProgram(
            shaders.compileShader(vshader,   GL_VERTEX_SHADER),
            shaders.compileShader(frag_shaded, GL_FRAGMENT_SHADER))
        self.wire_shader = shaders.compileProgram(
            shaders.compileShader(vshader,   GL_VERTEX_SHADER),
            shaders.compileShader(frag_wire, GL_FRAGMENT_SHADER))
        self.norm_shader = shaders.compileProgram(
            shaders.compileShader(vshader,   GL_VERTEX_SHADER),
            shaders.compileShader(frag_norm, GL_FRAGMENT_SHADER))
        self.path_shader = shaders.compileProgram(
            shaders.compileShader(path_shader,   GL_VERTEX_SHADER),
            shaders.compileShader(frag_path, GL_FRAGMENT_SHADER))
        self.cube_shader = shaders.compileProgram(
            shaders.compileShader(cube_vert,   GL_VERTEX_SHADER),
            shaders.compileShader(cube_frag, GL_FRAGMENT_SHADER))
            
################################################################################
    
    def init_GL(self):
        if glGetString(GL_VERSION) is None:
            return
        self.SetCurrent(self.context)
        self.compile_shaders()
        
        glEnable(GL_DEPTH_TEST)
        
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        
        self.init = True
    
################################################################################
    
    def load_mesh(self, mesh, reset=True):
        ''' Loads a ctypes array of floats into a vertex buffer object. '''
        
        self.vertex_vbo = vbo.VBO(mesh.vdata)

        self.index_vbo  = vbo.VBO(mesh.idata, target=GL_ELEMENT_ARRAY_BUFFER)
        self.bounds = mesh.X, mesh.Y, mesh.Z
        
        # Find useful parameters about the mesh.
        if reset:
            center = Vec3f(mesh.X.lower + mesh.X.upper,
                           mesh.Y.lower + mesh.Y.upper,
                           mesh.Z.lower + mesh.Z.upper) / 2
            self.center = Vec3f(center.x, center.y, center.z)
        

            radius = libtree.find_radius(mesh.vdata, len(mesh.vdata), center)
            self.scale = 1/radius
            self.scale *= 4
            
        self.Refresh()
    
    def load_mesh_vbos(self, vertices, indices, bounds):
        self.vertex_vbo = vertices
        self.index_vbo  = indices
        self.bounds = bounds
    
################################################################################
    
    def load_path(self, path):
        self.path_vbo = vbo.VBO(path.vdata)
        self.path     = path

################################################################################

    def load_cubes(self, cube_data):
        self.cube_vbo = vbo.VBO(cube_data)

################################################################################

    def draw_cubes(self, render=False):
    
        if not self.init:
            self.init_GL()
            self.update_viewport()
        if not self.init:
            return
        
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glColor3f(0.4, 1, 0.5)
        
        glPushMatrix()
        
        glTranslatef(0, 0, -5)
        glScalef(self.scale, self.scale, self.scale)
        glRotatef(-self.beta,  1, 0, 0)
        glRotatef(self.alpha, 0, 0, 1)
        glTranslatef(-self.center.x, -self.center.y, -self.center.z)

        current_shader = self.cube_shader
        shaders.glUseProgram(current_shader)
        
        self.cube_vbo.bind()

        # Find the positions of shader attributes
        attributes = {}
        for a in ['vertex_position', 'vertex_color']:
            attributes[a] = glGetAttribLocation(current_shader, a)
            glEnableVertexAttribArray(attributes[a])
        
        glVertexAttribPointer(
            attributes['vertex_position'], 3,
            GL_FLOAT, False, 6*4, self.cube_vbo)
            
        glVertexAttribPointer(
            attributes['vertex_color'], 3,
            GL_FLOAT, False, 6*4, self.cube_vbo+3*4)
        
        # Draw the triangles stored in the vbo
        glDrawArrays(GL_TRIANGLES, 0, len(self.cube_vbo))

        glDisableVertexAttribArray(attributes['vertex_position'])        
        glDisableVertexAttribArray(attributes['vertex_color'])

        self.cube_vbo.unbind()
        
        shaders.glUseProgram(0)
        
        glPopMatrix()
        
        if render:
            self.SwapBuffers()
        else:
            width, height = self.GetClientSize() 
            glReadBuffer(GL_BACK)
            glPixelStorei(GL_PACK_ALIGNMENT, 1) 
            return width, height, glReadPixels(2, 2, width,height, 
                                               GL_RGB, GL_UNSIGNED_BYTE)
        
################################################################################
    
    def draw(self):       
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        if self.index_vbo is None:
            if self.border is not None:
                self.draw_border()
            self.SwapBuffers()
            return
        
        if self.shader == 'wire':
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        else:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
            
        glPushMatrix()
        
        glTranslatef(0, 0, -5)
        glScalef(self.scale, self.scale, self.scale)
        glRotatef(-self.beta,  1, 0, 0)
        glRotatef(self.alpha, 0, 0, 1)
        glTranslatef(-self.center.x, -self.center.y, -self.center.z)
        
        current_shader = {'wire':   self.wire_shader,
                          'normal': self.norm_shader,
                          'plain':  self.plain_shader}[self.shader]
        
        shaders.glUseProgram(current_shader)
        self.vertex_vbo.bind()
        self.index_vbo.bind()

        # Find the positions of shader attributes
        attributes = {}
        for a in ['vertex_position', 'vertex_normal']:
            attributes[a] = glGetAttribLocation(current_shader, a)
            glEnableVertexAttribArray(attributes[a])
        
        glVertexAttribPointer(
            attributes['vertex_position'], # attribute index
            3,                                  # number of components per attribute
            GL_FLOAT,                           # data type of each component
            False,                              # Do not normalize components
            3*4*2,                              # stride length (in bytes)
            self.vertex_vbo)                    # Vertex buffer object
            
        glVertexAttribPointer(
            attributes['vertex_normal'], 3,
            GL_FLOAT, False, 3*4*2, self.vertex_vbo + 12)
        
        # Draw the triangles stored in the vbo
        glDrawElements(GL_TRIANGLES, len(self.index_vbo),
                       GL_UNSIGNED_INT, self.index_vbo)

        glDisableVertexAttribArray(attributes['vertex_normal'])        
        glDisableVertexAttribArray(attributes['vertex_position'])

        self.index_vbo.unbind()
        self.vertex_vbo.unbind()
        shaders.glUseProgram(0)
        
        self.draw_bounds()

        glPopMatrix()
        
        if self.path_vbo is not None:
            self.draw_paths()

        self.SwapBuffers()
        
################################################################################

    def draw_cube(self, X, Y, Z):
        glBegin(GL_QUADS)
        for z in (Z.upper, Z.lower):
            glVertex(X.lower, Y.lower, z)
            glVertex(X.lower, Y.upper, z)
            glVertex(X.upper, Y.upper, z)
            glVertex(X.upper, Y.lower, z)
        for y in (Y.upper, Y.lower):
            glVertex(X.lower, y, Z.lower)
            glVertex(X.lower, y, Z.upper)
            glVertex(X.upper, y, Z.upper)
            glVertex(X.upper, y, Z.lower)
        for x in (X.lower, X.upper):
            glVertex(x, Y.lower, Z.lower)
            glVertex(x, Y.lower, Z.upper)
            glVertex(x, Y.upper, Z.upper)
            glVertex(x, Y.upper, Z.lower)
        glEnd()

    def draw_bounds(self):
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)

        if self.border is None:
            glColor3f(0.4, 1, 0.5)
        else:
            glColor3f(self.border[0]/255.,
                      self.border[1]/255.,
                      self.border[2]/255.)
        self.draw_cube(*self.bounds)

        
################################################################################

    def draw_border(self):
        ''' Draws a colored border around the edge of the screen. '''
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
    
        w, h = self.Size
        glViewport(0, 0, w, h)
        
        aspect = w/float(h)

        if aspect > 1:
            xmin, xmax = -aspect, aspect
            ymin, ymax = -1, 1
        else:
            xmin, xmax = -1, 1
            ymin, ymax = -1/aspect, 1/aspect
            
        glBegin(GL_QUADS)    
        if self.border is None:
            glColor3f(0, 1, 0)
        else:
            glColor3f(self.border[0]/255.,
                      self.border[1]/255.,
                      self.border[2]/255.)
        
        glVertex(xmin, ymin, -1)
        glVertex(xmin, ymax, -1)
        glVertex(xmin+0.01, ymax, -1)
        glVertex(xmin+0.01, ymin, -1)

        glVertex(xmin, ymax, -1)
        glVertex(xmax, ymax, -1)
        glVertex(xmax, ymax-0.01, -1)
        glVertex(xmin, ymax-0.01, -1)

        glVertex(xmax-0.01, ymin, -1)
        glVertex(xmax-0.01, ymax, -1)
        glVertex(xmax, ymax, -1)
        glVertex(xmax, ymin, -1)

        glVertex(xmin, ymin+0.01, -1)
        glVertex(xmax, ymin+0.01, -1)
        glVertex(xmax, ymin, -1)
        glVertex(xmin, ymin, -1)

        glEnd()
        
################################################################################
        
    def draw_paths(self):
        glPushMatrix()
        
        glTranslatef(0, 0, -5)
        glScalef(self.scale, self.scale, self.scale)
        glRotatef(-self.beta,  1, 0, 0)
        glRotatef(self.alpha, 0, 0, 1)
        glScalef(1, -1, 1)
        glTranslatef(-(self.path.xmin+self.path.dx/2),
                     -(self.path.ymin+self.path.dy/2),
                     -(self.path.zmin+self.path.dz/2))
                     
        current_shader = self.path_shader
        shaders.glUseProgram(current_shader)
        
        self.path_vbo.bind()

        # Find the positions of shader attributes
        attributes = {}
        for a in ['vertex_position', 'vertex_color']:
            attributes[a] = glGetAttribLocation(current_shader, a)
            glEnableVertexAttribArray(attributes[a])
        
        glVertexAttribPointer(
            attributes['vertex_position'], 3,
            GL_FLOAT, False, 4*4, self.path_vbo)
            
        glVertexAttribPointer(
            attributes['vertex_color'], 1,
            GL_FLOAT, False, 4*4, self.path_vbo + 12)
        
        # Draw the triangles stored in the vbo
        glDrawArrays(GL_LINE_STRIP, 0, len(self.path_vbo))

        glDisableVertexAttribArray(attributes['vertex_position'])        
        glDisableVertexAttribArray(attributes['vertex_color'])

        self.path_vbo.unbind()
        
        shaders.glUseProgram(0)
        
        glPopMatrix()